from clp_ffi_py.ir.native import *
from clp_ffi_py.ir.query_builder import *
from clp_ffi_py.ir.readers import *

from typing import List

__all__: List[str] = [
    "Decoder",  # native
    "DecoderBuffer",  # native
    "FourByteEncoder",  # native
    "IncompleteStreamError",  # native
    "LogEvent",  # native
    "Metadata",  # native
    "Query",  # native
    "QueryBuilder",  # query_builder
    "ClpIrFileReader",  # readers
    "ClpIrStreamReader",  # readers
]



from pathlib import Path
from typing import List, Optional, Tuple

from test_ir.test_decoder import (
    TestCaseDecoderBase,
    TestCaseDecoderTimeRangeQueryBase,
    TestCaseDecoderTimeRangeWildcardQueryBase,
    TestCaseDecoderWildcardQueryBase,
)
from test_ir.test_utils import TestCLPBase

from clp_ffi_py.ir import (
    ClpIrFileReader,
    ClpIrStreamReader,
    IncompleteStreamError,
    LogEvent,
    Metadata,
    Query,
)


def read_log_stream(
    log_path: Path, query: Optional[Query], enable_compression: bool
) -> Tuple[Metadata, List[LogEvent]]:
    metadata: Metadata
    log_events: List[LogEvent] = []
    with open(str(log_path), "rb") as fin:
        reader = ClpIrStreamReader(fin, enable_compression=enable_compression)
        if None is query:
            for log_event in reader:
                log_events.append(log_event)
        else:
            for log_event in reader.search(query):
                log_events.append(log_event)
        metadata = reader.get_metadata()
        reader.close()
    return metadata, log_events


class TestCaseReaderBase(TestCaseDecoderBase):
    # override
    def _decode_log_stream(
        self, log_path: Path, query: Optional[Query]
    ) -> Tuple[Metadata, List[LogEvent]]:
        return read_log_stream(log_path, query, self.enable_compression)


class TestCaseReaderTimeRangeQueryBase(TestCaseDecoderTimeRangeQueryBase):
    # override
    def _decode_log_stream(
        self, log_path: Path, query: Optional[Query]
    ) -> Tuple[Metadata, List[LogEvent]]:
        return read_log_stream(log_path, query, self.enable_compression)


class TestCaseReaderWildcardQueryBase(TestCaseDecoderWildcardQueryBase):
    # override
    def _decode_log_stream(
        self, log_path: Path, query: Optional[Query]
    ) -> Tuple[Metadata, List[LogEvent]]:
        return read_log_stream(log_path, query, self.enable_compression)


class TestCaseReaderTimeRangeWildcardQueryBase(TestCaseDecoderTimeRangeWildcardQueryBase):
    # override
    def _decode_log_stream(
        self, log_path: Path, query: Optional[Query]
    ) -> Tuple[Metadata, List[LogEvent]]:
        return read_log_stream(log_path, query, self.enable_compression)


class TestCaseReaderDecompress(TestCaseReaderBase):
    """
    Tests stream reader methods against uncompressed IR stream.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = False
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderDecompressZstd(TestCaseReaderBase):
    """
    Tests stream reader against zstd compressed IR stream.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = False
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderTimeRangeQuery(TestCaseReaderTimeRangeQueryBase):
    """
    Tests stream reader against uncompressed IR stream with the query that
    specifies a search timestamp.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderTimeRangeQueryZstd(TestCaseReaderTimeRangeQueryBase):
    """
    Tests stream reader against zstd compressed IR stream with the query that
    specifies a search timestamp.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderWildcardQuery(TestCaseReaderWildcardQueryBase):
    """
    Tests stream reader against uncompressed IR stream with the query that
    specifies wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderWildcardQueryZstd(TestCaseReaderWildcardQueryBase):
    """
    Tests stream reader against zstd compressed IR stream with the query that
    specifies a wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderTimeRangeWildcardQuery(TestCaseReaderTimeRangeWildcardQueryBase):
    """
    Tests stream reader against uncompressed IR stream with the query that
    specifies both search time range and wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseReaderTimeRangeWildcardQueryZstd(TestCaseReaderTimeRangeWildcardQueryBase):
    """
    Tests stream reader against zstd compressed IR stream with the query that
    specifies both search time range and wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestIncompleteIRStream(TestCLPBase):
    """
    Tests on reading an incomplete stream.
    """

    test_src: Path = Path(__file__).resolve().parent / "test_data/incomplete_ir.log.zst"

    def test_incomplete_ir_stream_error(self) -> None:
        """
        Tests the reader against an incomplete IR file with
        `allow_incomplete_stream` disabled.
        """
        incomplete_stream_error_captured: bool = False
        other_exception_captured: bool = False
        log_counter: int = 0
        with ClpIrFileReader(TestIncompleteIRStream.test_src) as clp_reader:
            try:
                for _ in clp_reader.search(Query()):
                    log_counter += 1
            except IncompleteStreamError:
                incomplete_stream_error_captured = True
            except Exception:
                other_exception_captured = True
        self.assertTrue(
            incomplete_stream_error_captured, "Incomplete Stream Error is not properly set."
        )
        self.assertFalse(other_exception_captured, "No other exception should be set.")
        self.assertTrue(0 != log_counter, "No logs are decoded.")

    def test_allow_incomplete_ir_stream_error(self) -> None:
        """
        Tests the reader against an incomplete IR file with
        `allow_incomplete_stream` enabled.
        """
        incomplete_stream_error_captured: bool = False
        other_exception_captured: bool = False
        log_event: Optional[LogEvent] = None
        with ClpIrFileReader(
            TestIncompleteIRStream.test_src, allow_incomplete_stream=True
        ) as clp_reader:
            try:
                while None is not log_event:
                    log_event = clp_reader.read_next_log_event()
            except IncompleteStreamError:
                incomplete_stream_error_captured = True
            except Exception:
                other_exception_captured = True
        self.assertFalse(
            incomplete_stream_error_captured, "Incomplete Stream Error shouldn't be set."
        )
        self.assertFalse(other_exception_captured, "No other exception should be set.")
        self.assertTrue(None is log_event, "None is not reached.")



from typing import List, Optional

from test_ir.test_utils import TestCLPBase

from clp_ffi_py.ir import (
    Query,
    QueryBuilder,
    QueryBuilderException,
)
from clp_ffi_py.wildcard_query import WildcardQuery


class TestCaseQueryBuilder(TestCLPBase):
    """
    Class for testing clp_ffi_py.ir.QueryBuilder.
    """

    def test_init(self) -> None:
        """
        Tests the default initialized Query Builder and its behavior.
        """
        query_builder: QueryBuilder = QueryBuilder()
        empty_query: Query = Query()
        attributes_exception_captured: bool

        self._check_query(
            query_builder.build(),
            empty_query.get_search_time_lower_bound(),
            empty_query.get_search_time_upper_bound(),
            empty_query.get_wildcard_queries(),
            empty_query.get_search_time_termination_margin(),
        )

        attributes_exception_captured = False
        try:
            query_builder.search_time_lower_bound = 0  # type: ignore
        except AttributeError:
            attributes_exception_captured = True
        except Exception:
            pass
        self.assertTrue(
            attributes_exception_captured, "search time lower bound should be read-only"
        )

        attributes_exception_captured = False
        try:
            query_builder.search_time_upper_bound = 0  # type: ignore
        except AttributeError:
            attributes_exception_captured = True
        except Exception:
            pass
        self.assertTrue(
            attributes_exception_captured, "search time upper bound should be read-only"
        )

        attributes_exception_captured = False
        try:
            query_builder.search_time_termination_margin = 0  # type: ignore
        except AttributeError:
            attributes_exception_captured = True
        except Exception:
            pass
        self.assertTrue(
            attributes_exception_captured, "search time termination margin should be read-only"
        )

        query_builder.wildcard_queries.append(WildcardQuery(""))
        self.assertEqual(
            len(query_builder.wildcard_queries), 0, "The query list should be size of 0"
        )

    def test_set_value(self) -> None:
        """
        Tests QueryBuilder by building different Query objects.
        """
        search_time_lower_bound: int
        search_time_upper_bound: int
        search_time_termination_margin: int
        wildcard_queries: Optional[List[WildcardQuery]]
        query_builder: QueryBuilder = QueryBuilder()

        search_time_lower_bound = Query.default_search_time_lower_bound()
        search_time_upper_bound = Query.default_search_time_upper_bound()
        search_time_termination_margin = Query.default_search_time_termination_margin()
        wildcard_queries = None
        query_builder.set_search_time_lower_bound(search_time_lower_bound)
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            0,
        )

        search_time_upper_bound = 3270
        query_builder.set_search_time_upper_bound(search_time_upper_bound)
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        search_time_termination_margin = 123
        query_builder.set_search_time_termination_margin(search_time_termination_margin)
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        wildcard_queries = [WildcardQuery("aaa*aaa"), WildcardQuery("bbb*bbb", True)]
        for wildcard_query in wildcard_queries:
            query_builder.add_wildcard_query(
                wildcard_query.wildcard_query, wildcard_query.case_sensitive
            )
        extra_wildcard_queries = [WildcardQuery("ccc?ccc", True), WildcardQuery("ddd?ddd")]
        query_builder.add_wildcard_queries(extra_wildcard_queries)
        wildcard_queries.extend(extra_wildcard_queries)
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        search_time_lower_bound = 16384
        search_time_upper_bound = 16384
        query_builder = query_builder.set_search_time_lower_bound(
            search_time_lower_bound
        ).set_search_time_upper_bound(search_time_upper_bound)
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        search_time_termination_margin = 3670
        wildcard_query_string: str = "eee*?*eee"
        query_builder.set_search_time_termination_margin(
            search_time_termination_margin
        ).add_wildcard_query(wildcard_query_string)
        wildcard_queries.append(WildcardQuery(wildcard_query_string))
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        search_time_lower_bound = 12345
        search_time_upper_bound = 67890
        search_time_termination_margin = 0
        query_builder = (
            query_builder.reset_search_time_lower_bound()
            .set_search_time_lower_bound(search_time_lower_bound)
            .reset_search_time_termination_margin()
            .set_search_time_termination_margin(search_time_termination_margin)
            .reset_search_time_upper_bound()
            .set_search_time_upper_bound(search_time_upper_bound)
        )
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        query_builder.reset_wildcard_queries()
        query_builder.add_wildcard_query(wildcard_query_string)
        wildcard_queries = [WildcardQuery(wildcard_query_string)]
        self._check_query(
            query_builder.build(),
            search_time_lower_bound,
            search_time_upper_bound,
            wildcard_queries,
            search_time_termination_margin,
        )

        query_builder = query_builder.reset()
        self._check_query(
            query_builder.build(),
            Query.default_search_time_lower_bound(),
            Query.default_search_time_upper_bound(),
            None,
            0,
        )

    def test_exception(self) -> None:
        """
        Tests whether QueryBuilderException is triggered as expected.
        """
        query_builder: QueryBuilder = QueryBuilder()
        search_time_lower_bound: int = 3270
        search_time_upper_bound: int = 3190
        query_builder.set_search_time_lower_bound(search_time_lower_bound)
        query_builder.set_search_time_upper_bound(search_time_upper_bound)

        query_builder_exception_captured: bool = False
        try:
            query_builder.build()
        except QueryBuilderException:
            query_builder_exception_captured = True
        except Exception:
            pass
        self.assertTrue(
            query_builder_exception_captured,
            "QueryBuilderException is not triggered when the search time lower bound exceeds the"
            " search time upper bound",
        )



# -- Project information -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "clp-ffi-py"

# -- General configuration -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_design",
    "sphinx_tippy",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx_copybutton",
]

templates_path = ["_templates"]

# -- HTML output options -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_favicon = "https://docs.yscope.com/_static/favicon.ico"
html_title = "clp-ffi-py"
html_show_copyright = False
html_static_path = ["../src/_static"]

html_theme = "furo"

# -- Furo theme options --------------------------------------------------------
# https://furo.readthedocs.io/customisation/index.html
html_theme_options = {
    "source_repository": "https://github.com/y-scope/clp-ffi-py/",
    "source_branch": "main",
    "source_directory": "docs",
}

# -- MyST extensions -----------------------------------------------------------
# https://myst-parser.readthedocs.io/en/stable/syntax/optional.html
myst_enable_extensions = [
    "colon_fence",
]

# -- Sphinx autodoc options ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

autoclass_content = "class"
autodoc_class_signature = "separated"
autodoc_typehints = "description"



from pathlib import Path
from sys import stderr
from types import TracebackType
from typing import Generator, IO, Iterator, Optional, Type, Union

from zstandard import ZstdDecompressionReader, ZstdDecompressor

from clp_ffi_py.ir.native import Decoder, DecoderBuffer, LogEvent, Metadata, Query


class ClpIrStreamReader(Iterator[LogEvent]):
    """
    This class represents a stream reader used to read/decode encoded log events
    from a CLP IR stream. It also provides method(s) to instantiate a log event
    generator with a customized search query.

    :param istream: Input stream that contains encoded CLP IR.
    :param decoder_buffer_size: Initial size of the decoder buffer.
    :param enable_compression: A flag indicating whether the istream is
        compressed using `zstd`.
    :param allow_incomplete_stream: If set to `True`, an incomplete CLP IR
        stream is not treated as an error. Instead, encountering such a stream
        is seen as reaching its end without raising any exceptions.
    """

    DEFAULT_DECODER_BUFFER_SIZE: int = 65536

    def __init__(
        self,
        istream: IO[bytes],
        decoder_buffer_size: int = DEFAULT_DECODER_BUFFER_SIZE,
        enable_compression: bool = True,
        allow_incomplete_stream: bool = False,
    ):
        self.__istream: Union[IO[bytes], ZstdDecompressionReader]
        if enable_compression:
            dctx: ZstdDecompressor = ZstdDecompressor()
            self.__istream = dctx.stream_reader(istream, read_across_frames=True)
        else:
            self.__istream = istream
        self._decoder_buffer: DecoderBuffer = DecoderBuffer(self.__istream, decoder_buffer_size)
        self._metadata: Optional[Metadata] = None
        self._allow_incomplete_stream: bool = allow_incomplete_stream

    def read_next_log_event(self) -> Optional[LogEvent]:
        """
        Reads and decodes the next encoded log event from the IR stream.

        :return:
            - Next unread log event represented as an instance of LogEvent.
            - None if the end of IR stream is reached.
        :raise Exception:
            If :meth:`~clp_ffi_py.ir.native.Decoder.decode_next_log_event`
            fails.
        """
        return Decoder.decode_next_log_event(
            self._decoder_buffer, allow_incomplete_stream=self._allow_incomplete_stream
        )

    def read_preamble(self) -> None:
        """
        Try to decode the preamble and set `metadata`. If `metadata` has been
        set already, it will instantly return. It is separated from `__init__`
        so that the input stream does not need to be readable on a reader's
        construction, but until the user starts to iterate logs.

        :raise Exception:
            If :meth:`~clp_ffi_py.ir.native.Decoder.decode_preamble` fails.
        """
        if self.has_metadata():
            return
        self._metadata = Decoder.decode_preamble(self._decoder_buffer)

    def get_metadata(self) -> Metadata:
        if None is self._metadata:
            raise RuntimeError("The metadata has not been successfully decoded yet.")
        return self._metadata

    def has_metadata(self) -> bool:
        return None is not self._metadata

    def search(self, query: Query) -> Generator[LogEvent, None, None]:
        """
        Searches and yields log events that match a specific search query.

        :param query: The input query object used to match log events. Check the
            document of :class:`~clp_ffi_py.ir.Query` for more details.
        :yield: The next unread encoded log event that matches the given search
            query from the IR stream.
        """
        if False is self.has_metadata():
            self.read_preamble()
        while True:
            log_event: Optional[LogEvent] = Decoder.decode_next_log_event(
                self._decoder_buffer,
                query=query,
                allow_incomplete_stream=self._allow_incomplete_stream,
            )
            if None is log_event:
                break
            yield log_event

    def close(self) -> None:
        self.__istream.close()

    def __iter__(self) -> "ClpIrStreamReader":
        if False is self.has_metadata():
            self.read_preamble()
        return self

    def __enter__(self) -> "ClpIrStreamReader":
        if False is self.has_metadata():
            self.read_preamble()
        return self

    def __next__(self) -> LogEvent:
        next_log_event: Optional[LogEvent] = self.read_next_log_event()
        if None is next_log_event:
            raise StopIteration
        return next_log_event

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()


class ClpIrFileReader(ClpIrStreamReader):
    """
    Wrapper class of `ClpIrStreamReader` that calls `open` for convenience.
    """

    def __init__(
        self,
        fpath: Path,
        decoder_buffer_size: int = ClpIrStreamReader.DEFAULT_DECODER_BUFFER_SIZE,
        enable_compression: bool = True,
        allow_incomplete_stream: bool = False,
    ):
        self._path: Path = fpath
        super().__init__(
            open(fpath, "rb"),
            decoder_buffer_size=decoder_buffer_size,
            enable_compression=enable_compression,
            allow_incomplete_stream=allow_incomplete_stream,
        )

    def dump(self, ostream: IO[str] = stderr) -> None:
        for log_event in self:
            ostream.write(str(log_event))



import os
import platform
import sys
import toml
from setuptools import setup, Extension
from typing import Any, Dict, List, Optional

ir_native: Extension = Extension(
    name="clp_ffi_py.ir.native",
    language="c++",
    include_dirs=[
        "src",
        "src/clp/components/core/submodules",
    ],
    sources=[
        "src/clp/components/core/src/BufferReader.cpp",
        "src/clp/components/core/src/ffi/ir_stream/decoding_methods.cpp",
        "src/clp/components/core/src/ffi/ir_stream/encoding_methods.cpp",
        "src/clp/components/core/src/ffi/encoding_methods.cpp",
        "src/clp/components/core/src/ir/parsing.cpp",
        "src/clp/components/core/src/ReaderInterface.cpp",
        "src/clp/components/core/src/string_utils.cpp",
        "src/clp/components/core/src/TraceableException.cpp",
        "src/clp_ffi_py/ir/native/decoding_methods.cpp",
        "src/clp_ffi_py/ir/native/encoding_methods.cpp",
        "src/clp_ffi_py/ir/native/Metadata.cpp",
        "src/clp_ffi_py/ir/native/PyDecoder.cpp",
        "src/clp_ffi_py/ir/native/PyDecoderBuffer.cpp",
        "src/clp_ffi_py/ir/native/PyFourByteEncoder.cpp",
        "src/clp_ffi_py/ir/native/PyLogEvent.cpp",
        "src/clp_ffi_py/ir/native/PyMetadata.cpp",
        "src/clp_ffi_py/ir/native/PyQuery.cpp",
        "src/clp_ffi_py/ir/native/Query.cpp",
        "src/clp_ffi_py/modules/ir_native.cpp",
        "src/clp_ffi_py/Py_utils.cpp",
        "src/clp_ffi_py/utils.cpp",
    ],
    extra_compile_args=[
        "-std=c++20",
        "-O3",
    ],
    define_macros=[("SOURCE_PATH_SIZE", str(len(os.path.abspath("./src/clp/components/core"))))],
)

if "__main__" == __name__:
    try:
        if "Darwin" == platform.system():
            target: Optional[str] = os.environ.get("MACOSX_DEPLOYMENT_TARGET")
            if None is target or float(target) < 10.15:
                os.environ["MACOSX_DEPLOYMENT_TARGET"] = "10.15"

        config: Dict[str, Any] = toml.load("pyproject.toml")
        version: Optional[str] = config.get("project", {}).get("version", None)
        if None is version:
            sys.exit("Error: The version number was not found in pyproject.toml")

        project_name: str = "clp_ffi_py"
        description: str = "CLP FFI Python Interface"
        extension_modules: List[Extension] = [ir_native]
        if (3, 7) > sys.version_info:
            # For Python3.6, setuptools doesn't automatically include submodules
            # and .pyi/.type files, so we need to explicitly specify the
            # packages and the files to include per package (package_data).
            packages: List[str] = ["clp_ffi_py", "clp_ffi_py.ir"]
            data_to_include: List[str] = ["*.py", "*.pyi", "*.typed"]
            package_data: Dict[str, List[str]] = {package: data_to_include for package in packages}
            setup(
                name=project_name,
                description=description,
                ext_modules=extension_modules,
                packages=packages,
                package_data=package_data,
                version=version,
            )
        else:
            setup(
                name=project_name,
                description=description,
                ext_modules=extension_modules,
                packages=["clp_ffi_py"],
                version=version,
            )

    except Exception as e:
        sys.exit(f"Error: {e}")



from test_ir.test_utils import TestCLPBase

from clp_ffi_py.ir import FourByteEncoder


class TestCaseFourByteEncoder(TestCLPBase):
    """
    Class for testing clp_ffi_py.ir.FourByteEncoder.

    The actual functionality should also be covered by the unittest of CLP
    Python logging library.
    TODO: When the decoder is implemented, add some more tests to ensure the
    encoded bytes can be successfully decoded to recover the original log event.
    """

    def test_init(self) -> None:
        type_error_exception_captured: bool = False
        four_byte_encoder: FourByteEncoder
        try:
            four_byte_encoder = FourByteEncoder()  # noqa
        except TypeError:
            type_error_exception_captured = True
        self.assertEqual(
            type_error_exception_captured, True, "FourByteEncoder should be non-instantiable."
        )

    def test_encoding_methods_consistency(self) -> None:
        """
        This test checks if the result of encode_message_and_timestamp_delta is
        consistent with the combination of encode_message and
        encode_timestamp_delta.
        """
        timestamp_delta: int = -3190
        log_message: str = "This is a test message: Do NOT Reply!"
        encoded_message_and_ts_delta: bytearray = (
            FourByteEncoder.encode_message_and_timestamp_delta(
                timestamp_delta, log_message.encode()
            )
        )
        encoded_message: bytearray = FourByteEncoder.encode_message(log_message.encode())
        encoded_ts_delta: bytearray = FourByteEncoder.encode_timestamp_delta(timestamp_delta)
        self.assertEqual(encoded_message_and_ts_delta, encoded_message + encoded_ts_delta)
